/*
 * Copyright (c) 2006 dean gaudet <dean@arctic.org>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

/* $Id: circular.c,v 1.3 2006-07-19 17:45:10 dean Exp $ */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <signal.h>

#define BLOCK_SIZE	(4096)

static const char *progname;
static int flush_it;

static void safe_write(int f, const void *b, size_t l)
{
    int rc;

    while (l) {
	rc = write(f, b, l);
	if (rc == -1) {
	    if (errno != EINTR) {
		fprintf(stderr, "%s: write error: %s\n", progname,
		    strerror(errno));
		exit(1);
	    }
	}
	if (rc > 0) {
	    l -= rc;
	    b = (void *)((char *)b + rc);
	}
    }
}

static void sig_handler(int sig)
{
    flush_it = sig;
}

int main(int argc, char **argv)
{
    char *buffer;
    int num_blocks;
    char *cur_ptr;
    char *next_block;
    char *top_of_buffer;
    int rc;
    int overflowed;
    struct sigaction sa;
    char *p;

    progname = strrchr(argv[0], '/');
    if (progname == NULL) progname = argv[0];
    if (argc != 2) {
	fprintf(stderr, "usage: %s bufsize_in_MiB\n", progname);
	exit(1);
    }

    sigemptyset(&sa.sa_mask);
    sa.sa_flags = 0;
    sa.sa_handler = sig_handler;
    if (sigaction(SIGINT, &sa, NULL) < 0) {
	fprintf(stderr, "%s: error installing SIGINT handler: %s\n", progname,
		strerror(errno));
	exit(1);
    }
    if (sigaction(SIGQUIT, &sa, NULL) < 0) {
	fprintf(stderr, "%s: error installing SIGQUIT handler: %s\n", progname,
		strerror(errno));
	exit(1);
    }

    num_blocks = strtoul(argv[1], &p, 0);
    if (num_blocks == 0 || *p) {
	fprintf(stderr, "%s: bufsize must be an integer > 0\n", progname);
	exit(1);
    }
    num_blocks *= (1024*1024)/BLOCK_SIZE;

    if ((buffer = malloc((num_blocks + 1) * BLOCK_SIZE)) == NULL) {
	fprintf(stderr, "%s: error allocating buffer\n", progname);
	exit(1);
    }

    /* align it to a BLOCK_SIZE boundary */
    buffer = (char *)((((unsigned long)buffer) + BLOCK_SIZE - 1)
			& ~(BLOCK_SIZE - 1));

    top_of_buffer = buffer + num_blocks * BLOCK_SIZE;

    /* indicate that we haven't overflowed the buffer yet */
    overflowed = 0;
    for (;;) {
	/* start at beginning of buffer */
	cur_ptr = buffer;
	next_block = buffer + BLOCK_SIZE;
	for (;;) {
	    if (flush_it) {
		rc = -1;
		errno = EINTR;
	    }
	    else {
		rc = read(STDIN_FILENO, cur_ptr, next_block - cur_ptr);
	    }
	    if (rc <= 0) {
		int save_errno = errno;

		if (rc == -1 && save_errno != EINTR) {
		    fprintf(stderr, "%s: read error: %s\n",
			    progname, strerror(save_errno));
		}
		if (overflowed) {
		    safe_write(STDOUT_FILENO, cur_ptr, top_of_buffer - cur_ptr);
		}
		safe_write(STDOUT_FILENO, buffer, cur_ptr - buffer);
		if (rc == -1 && save_errno != EINTR) {
		    exit(1);
		}
		if (rc == 0 || flush_it == SIGQUIT) {
		    exit(0);
		}
		flush_it = 0;
		overflowed = 0;
		break;
	    }
	    cur_ptr += rc;
	    if (cur_ptr == next_block) {
		next_block += BLOCK_SIZE;
		if (next_block > top_of_buffer) {
		    /* we have overflowed the buffer */
		    overflowed = 1;
		    break;
		}
	    }
	}
    }
    return 0;
}
